--- Guilder: Raid Manager Module
-- This module allows the user to manage the members' Raid
-- @class file
-- @name Modules/Raid.lua
-- @release 2.1.x

--[[
   Copyright 2009 Christo Kotze (onaforeignshore)

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

	 http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
]]

if select(6, GetAddOnInfo("Guilder_" .. (debugstack():match("[o%.][d%.][u%.]les\\(.-)\\") or ""))) ~= "MISSING" then return end

local Guilder = _G.Guilder
if not Guilder then
  error("Guilder_Raid requires Guilder")
end
local L = LibStub("AceLocale-3.0"):GetLocale("Guilder")

local DEBUG = Guilder.DEBUG
local expect = Guilder.expect

-- default settings for the addon -------------------------------------------

local DATABASE_DEFAULTS = {
	announceRoll = false,
	includeOffline = false,
	allowAssist = false,
	findIDsBy = "instance",
	difficulty = "heroic",
	which = ""
}

-- declare the module -------------------------------------------------------

local Guilder_Raid = Guilder:NewModule("Raid", "AceEvent-3.0", "AceHook-3.0")
Guilder_Raid:SetName(L["Raid Wizard"])
Guilder_Raid:SetDescription(L["Various Raid related utilities."])
Guilder_Raid:SetDefaults(DATABASE_DEFAULTS)
if DEBUG then
	Guilder_Raid:SetIcon([[Interface\AddOns\Guilder\Modules\Raid\Raid]])
else
	Guilder_Raid:SetIcon([[Interface\AddOns\Guilder_Raid\Raid]])
end

-- localise global variables for faster access ------------------------------

local RequestRaidInfo = RequestRaidInfo
local UnitIsPlayer = UnitIsPlayer

-- local variables ----------------------------------------------------------

-- database --
local instance2player = {}
local player2instance = {}

--[===[@debug@
Guilder.i2p = instance2player
Guilder.p2i = player2instance
--@end-debug@]===]

-- Standard Module Methods --------------------------------------------------

-- Is called when the module is loaded
-- @usage internal use only
function Guilder_Raid:OnModuleLoaded(module)
	if not self.db.profile.enabled then return end
	local id = module.id
	self.db.profile.which = ""
end

-- Is called when the module is enabled
-- @usage internal use only
function Guilder_Raid:OnEnable()
	-- register events for when the group changes
	LibGroup.RegisterCallback(self, "AssistAdded", "RefreshConfig")
	LibGroup.RegisterCallback(self, "AssistRemoved", "RefreshConfig")
	LibGroup.RegisterCallback(self, "LeaderChanged", "RefreshConfig")
	LibGroup.RegisterCallback(self, "MasterLooterChanged", "RefreshConfig")
	-- register instance events
	self:RegisterEvent("UPDATE_INSTANCE_INFO")
	RequestRaidInfo() -- triggers UPDATE_INSTANCE_INFO
end

-- Is called when the module is disabled
-- @usage internal use only
function Guilder_Raid:OnDisable()
	LibGroup.UnregisterAll()
	self:UnregisterAll()
end

-- Is called when the user opens the configuration screen
-- @usage internal use only
function Guilder_Raid:UpdateConfigOptions()
	if Guilder.configLoaded then
		-- set the right options for instance IDs
		Guilder_Raid:UpdateFindIDsBy()
		-- enable raid roll button if we're an assistant, leader or masterlooter
		self.options.div.args.raidRoll.disabled = not (LibGroup:IsInGroup() and (LibGroup:GetRank() > 0 or LibGroup:IsMasterLooter()))
		-- create the ready check / ready check request options
		if Guilder:IsInRaid() then
			if LibGroup:GetRank() == 2 then
				self.options.div.args.h2 = {
					type = "header",
					name = L["Ready Check Request"],
					order = 5,
				}
				self.options.div.args.readyCheck = {
					type = "execute",
					name = L["Ready Check"],
					desc = L["Do a Ready Check"],
					order = 6,
					func = function() self:ReadyCheck() end
				}
				if Guilder:IsInRaid() then
					self.options.div.args.allowAssist = {
						type = "toggle",
						name = L["Allow Assist Requests"],
						desc = L["Automatically do a Ready Check when Raid Assistant requests one?"],
						order = 7,
						get = function() return self.db.profile.allowAssist end,
						set = function(info, v) self.db.profile.allowAssist = v end
					}
				else
					self.options.div.args.allowAssist = nil
				end
			elseif LibGroup:GetRank() == 1 then
				self.options.div.args.h2 = {
					type = "header",
					name = L["Ready Check Request"],
					order = 5,
				}
				self.options.div.args.readyCheck = {
					type = "execute",
					name = L["Request Ready Check"],
					desc = L["Request a Ready Check from the group leader"],
					order = 6,
					func = function() self:ReadyCheck() end
				}
				self.options.div.args.allowAssist = nil
			else
				self.options.div.args.h2 = nil
				self.options.div.args.readyCheck = nil
				self.options.div.args.allowAssist = nil
			end
		end
	end
end

-- Is called when our instance info is updated
-- @usage internal use only
function Guilder_Raid:UPDATE_INSTANCE_INFO()
	local iAmSavedTo = {}
	iAmSavedTo["heroic"] = iAmSavedTo["heroic"] or {}
	iAmSavedTo["normal"] = iAmSavedTo["normal"] or {}
	for i = 1, GetNumSavedInstances() do
		-- name, id, reset, difficulty, locked, extended, instanceIDMostSig, isRaid, maxPlayers, difficultyName = GetSavedInstanceInfo(index)
		local name, id, _, difficulty, locked, extended, _, isRaid, maxPlayers, _ = GetSavedInstanceInfo(i)
		if locked or extended then
			if isRaid then name = name.." ("..maxPlayers..")" end
			if difficulty > 1 then
				iAmSavedTo["heroic"][name] = id
			else
				iAmSavedTo["normal"][name] = id
			end
		end
	end
	Guilder:SendComm(self.id, "GUILD", myName, "URII", iAmSavedTo)
end

-- Is called when the module receives a message
-- @usage internal use only
function Guilder_Raid:CommHandler(distribution, sender, command, ...)
	-- if module is disabled, do nothing
	if not self.db.profile.enabled then
		if command == "URRC" then
			Guilder:ShowMessage(L["%s tried to request a Ready Check, but Raid Wizard was disabled."]:format(sender))
			Guilder:SendComm(self.id, "WHISPER", sender, "RRRC", false)
		end
		return
	end
	-- only handle comms if they are not from yourself, except for raid instance info
	if sender == myName and command ~= "URII" then return end
	local t = {}
	local n = select("#", ...)
	-- Query: Raid Instance Info
	if command == "QRII" then
		RequestRaidInfo() -- triggers UPDATE_INSTANCE_INFO
	end
	-- Update: Raid Instance Info
	if command == "URII" then
		-- get the first parameter
		t = select(1, ...)
		-- check if the message contains a table, otherwise ignore it
		if n ~= 1 or type(t) ~= "table" then
			--[===[@debug@
			if self.db.profile.showDebug then
				self:ShowMessage("URII expected a table!")
			end
			--@end-debug@]===]
			return
		end
		-- process the incoming instance info update
		for difficulty, diffTable in pairs(t) do
			for instanceName, ID in pairs(diffTable) do
				instance2player[difficulty] = instance2player[difficulty] or {}
				instance2player[difficulty][instanceName] = instance2player[difficulty][instanceName] or {}
				instance2player[difficulty][instanceName][sender] = ID
				player2instance[difficulty] = player2instance[difficulty] or {}
				player2instance[difficulty][sender] = player2instance[difficulty][sender] or {}
				player2instance[difficulty][sender][instanceName] = ID
			end
		end
	end
	-- Update: Request Ready Check
	if command == "URRC" then
		Guilder:ShowMessage(L["%s requested a Ready Check"]:format(sender))
		Guilder:SendComm(self.id, "WHISPER", sender, "RRRC", true)
		-- do a ready check if allowed
		if self.db.profile.allowAssist() then
			DoReadyCheck()
		end
	end
	-- Response: Request Ready Check
	if command == "RRRC" then
		if n == 1 then
			if select(1, ...) then
				Guilder:ShowMessage(L["%s has received your Ready Check request"]:format(sender))
			else
				Guilder:ShowMessage(L["%s's Raid Wizard is disabled! Ready Check request has been declined."]:format(sender))
			end
		end
	end
end

function Guilder_Raid:RefreshConfig()
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
end

-- General functions --------------------------------------------------------

function Guilder_Raid:RaidRoll()
	if not self.db.profile.enabled then
		Guilder:ShowMessage(L["%s is disabled"]:format("|cff00c0ff"..L["Raid Wizard"].."|r"))
		return
	end
	if self.db.profile.announceRoll then
		Guilder:SendMessage(L["%s is Raid Rolling! Check the raid channel for the result."]:format(myName), "RAID_WARNING")
	end
	local t = {}
	for k, _ in LibGroup:Iterate(self.db.profile.includeOffline) do
		t[#t + 1] = k
	end
	local r = random(#t)
	local m
	for i = 1, #t, 5 do
		m = i..". "..t[i]
		if t[i + 1] then m = m.."   "..(i + 1)..". "..t[i + 1] end
		if t[i + 2] then m = m.."   "..(i + 2)..". "..t[i + 2] end
		if t[i + 3] then m = m.."   "..(i + 3)..". "..t[i + 3] end
		if t[i + 4] then m = m.."   "..(i + 4)..". "..t[i + 4] end
		Guilder:SendMessage(m, "RAID")
	end
	Guilder:SendMessage(L["Rolled %s. %s wins!"]:format(r, t[r]), "RAID")
end

function Guilder_Raid:ReadyCheck()
	if LibGroup:GetRank() == 2 then
		DoReadyCheck()
	else
		Guilder:SendComm(self.id, "WHISPER", LibGroup:GetGroupLeader(), "URRC")
	end
end

function Guilder_Raid:GetPlayers()
	local t = {}
	if player2instance[self.db.profile.difficulty] then
		for k, _ in pairs(player2instance[self.db.profile.difficulty]) do
			t[k] = "|cff"..LibGuild:GetClassHexColor(k)..k.."|r"
		end
	end
	return t
end

function Guilder_Raid:GetInstances()
	local t = {}
	if instance2player[self.db.profile.difficulty] then
		for k, _ in pairs(instance2player[self.db.profile.difficulty]) do
			t[k] = k
		end
	end
	return t
end

function Guilder_Raid:ShowInstanceInfo()
	if not self.db.profile.enabled then
		Guilder:ShowMessage(L["%s is disabled"]:format("|cff00c0ff"..L["Raid Wizard"].."|r"))
		return
	end
	if self.db.profile.which == "" then
		if self.db.profile.findIDsBy == "player" then
			Guilder:ShowMessage("|cff00c0ff"..L["Raid Wizard"].."|r - "..L["You have not selected a player!"])
		else
			Guilder:ShowMessage("|cff00c0ff"..L["Raid Wizard"].."|r - "..L["You have not selected an instance!"])
		end
		return
	end
	local t = {}
	local which = self.db.profile.which
	local dif = (self.db.profile.difficulty == "heroic") and L["Heroic"] or L["Normal"]
	if self.db.profile.findIDsBy == "player" then
		if player2instance[self.db.profile.difficulty][which] then
			Guilder:ShowMessage(L["IDs for %s"]:format("|cff"..LibGuild:GetClassHexColor(which)..which.."|r ("..dif..")"))
			for instanceName, ID in pairs(player2instance[self.db.profile.difficulty][which]) do
				Guilder:ShowMessage(instanceName.." - "..ID)
			end
		end
	else
		if instance2player[self.db.profile.difficulty][which] then
			Guilder:ShowMessage(L["IDs for %s"]:format("|cff33ff99"..which.."|r ("..dif..")"))
			for player, ID in pairs(instance2player[self.db.profile.difficulty][which]) do
				Guilder:ShowMessage(player.." - "..ID)
			end
		end
	end
end

function Guilder_Raid:UpdateFindIDsBy()
	if self.db.profile.findIDsBy == "player" then
		self.options.div.args.find.name = L["Player"]
		self.options.div.args.find.desc = L["Which player's IDs do you want to see?"]
		self.options.div.args.find.values = self:GetPlayers()
		self.db.profile.which = ""
	else
		self.options.div.args.find.name = L["Instance"]
		self.options.div.args.find.desc = L["Which instance's IDs do you want to see?"]
		self.options.div.args.find.values = self:GetInstances()
		self.db.profile.which = ""
	end
	LibStub("AceConfigRegistry-3.0"):NotifyChange("Guilder")
end
